# -*- ruby -*-

require 'dl.o'


module DL
  module Compat
    class PtrData
      def initialize(addr, size = nil, free = nil)
	@addr = addr
	@size = size
	@free = free
        @struct = nil
        @ctype = nil
        @struct = nil
        @mems = nil
      end

      def to_ptr
	self
      end

      def to_i
	@addr
      end

      def to_s(size = nil)
	if( size )
	  ::DL.dlwrap(@addr, size)
	else
	  ::DL.dlwrap(@addr, @size)
	end
      end

      def define_data_type(ctype, ty=nil, *mems)
        if( ty )
          @ctype  = ctype
          @struct = ty
          @mems  = mems
        else
          @ctype = :unknown
          @struct = nil
          @mems = nil
        end
      end

      def struct!(ty=nil, *mems)
        define_data_type(:struct, ty, mems)
      end

      def union!(ty=nil, *mems)
        define_data_type(:union, ty, mems)
      end

      def [](pos, size=nil)
      end

      def []=(pos, arg1, arg2=nil)
        if( arg2 )
        else
        end
      end

      private

      def align(addr, align)
        d = addr % align
        if( d == 0 )
          addr
        else
          addr + (align - d)
        end
      end
    end

    class Symbol
      NO_ENC = NO_DEC = proc{|x| x}

      def initialize(addr, name, type)
	if( !type || type.size == 0 )
	  type = '0'
	end
	retty = nil
	@rdec = NO_DEC
	case type[0]
	when ?0
	  retty = DL::TYPE_VOID
	when ?h, ?H
	  retty = DL::TYPE_SHORT
	when ?s, ?S
	  retty = DL::TYPE_VOIDP
	  @rdec = proc{|x| ::DL.dlwrap(x)}
	when ?p, ?P
	  retty = DL::TYPE_VOIDP
	  @rdec = proc{|x| PtrData.new(x)}
	when ?i, ?I
	  retty = DL::TYPE_INT
	when ?l, ?L
	  retty = DL::TYPE_LONG
	when ?f, ?F
	  retty = DL::TYPE_FLOAT
	when ?d, ?D
	  retty = DL::TYPE_DOUBLE
	end
	@cfunc    = ::DL::CFunc.new(addr, retty, name)
	@template = nil
	@enc      = NO_ENC
	@dec      = NO_DEC
	parse_type(type[1 .. (-1)])
      end
      
      def parse_type(type)
	@template = ""
	@enc      = NO_ENC
	@dec      = NO_DEC
	i = 0
	type.each_byte{|c|
	  case c
	  when ?C
	    t = 'c'
	    enc = dec = nil
	  when ?H
	    t = 's!'
	    enc = dec = nil
	  when ?I
	    t = 'i!'
	    enc = dec = nil
	  when ?L
	    t = 'l!'
	    enc = dec = nil
	  when ?F
	    t = 'f'
	    enc = dec = nil
	  when ?D
	    t = 'd'
	    enc = dec = nil
	  when ?c
	    t = 'p'
	    enc = proc{|x| [x].pack('c')}
	    dec = proc{|x| x.unpack('c')[0]}
	  when ?h
	    t = 'P'
	    enc = proc{|x| [x].pack('s!')}
	    dec = proc{|x| x.unpack('s!')[0]}
	  when ?i
	    t = 'P'
	    enc = proc{|x| [x].pack('i!')}
	    dec = proc{|x| x.unpack('i!')[0]}
	  when ?l
	    t = 'P'
	    enc = proc{|x| [x].pack('l!')}
	    dec = proc{|x| x.unpack('l!')[0]}
	  when ?f
	    t = 'P'
	    enc = proc{|x| [x].pack('f')}
	    dec = proc{|x| x.unpack('f')[0]}
	  when ?d
	    t = 'P'
	    enc = proc{|x| [x].pack('d')}
	    dec = proc{|x| x.unpack('d')[0]}
	  when ?S
	    t = 'p'
	    enc = dec = nil
	  when ?s
	    t = 'p'
	    enc = proc{|x| x.dup }
	    dec = nil
	  when ?P, ?p
	    t = 'l!'
	    enc = proc{|x| x.to_ptr.to_i}
	    dec = proc{|x| PtrData.new(x)}
	  else
	    raise(::DL::TypeError, "unknown type '#{c.chr}'")
	  end
	  proc{
	    j = i
	    a,b,e,d = enc,dec,@enc,@dec
	    if( a )
	      @enc = proc{|x| x[j] = a.call(x[j]); e.call(x)}
	    end
	    if( b )
	      @dec = proc{|x| x[j] = b.call(x[j]); d.call(x)}
	    end
	  }.call
	  @template.concat(t)
	  i += 1
	}
      end

      def call(*args)
	orig_args = args
	args = @enc.call(args)
	args = args.pack(@template).unpack("l!*")
	ret = @cfunc.call(args)
	args = @dec.call(args)
	[@rdec.call(ret), orig_args]
      end
      
      def [](*args)
	call(*args)
      end
    end

    class Handle
      def initialize(path, flag = nil, &block)
	flag ||= RTLD_LAZY | RTLD_GLOBAL
	@handle = ::DL::Handle.new(path, flag, &block)
      end

      def sym(name, type=nil)
	if( type )
	  cfunc = @handle.sym(name)
	else
	  cfunc = @handle.sym(name)
	  type  = "0"
	end
	Symbol.new(cfunc, name, type)
      end
      
      def [](name, type=nil)
	sym(name, type)
      end

      def close
	@handle.close
      end

      def to_i
	@handle.to_i
      end

      def to_ptr
	PtrData.new(@handle.to_i)
      end
    end
  end

  alias _dlopen dlopen
  def dlopen(*args)
    ::DL::Compat::Handle.new(*args)
  end
  module_function :dlopen

  include Compat
end

class String
  def to_ptr
    DL::Compat::PtrData.new([self].pack("p").unpack("l!")[0], self.size, nil)
  end
end
