require 'dl/compat/dl'
require 'dl/callback'

LIBC_SO = ARGV[0] || "/lib/libc.so.6"
LIBM_SO = ARGV[1] || "/lib/libm.so.6"

$c = DL.dlopen(LIBC_SO)
$m = DL.dlopen(LIBM_SO)
$num = 0
$err = 0

def test(phase, x1, x2)
  $num += 1
  print(phase, ": ")
  if( x2 === x1 )
    print("ok\n")
  else
    print("fail\n")
    $err += 1
  end
end

include DL
include Math

x = $m['sin','DD'][3.14/2]
test("sin(3.14/2)", x[0], Math.sin(3.14/2))

x = $m['sin','DD'][-3.14/2]
test("sin(-3.14/2)", x[0], Math.sin(-3.14/2))

x = $c['strlen', 'IS']['abc']
test("strlen('abc')", x[0], 'abc'.size)

x = $c['strlen', 'IP']['abc']
test("strlen('abc')", x[0], 'abc'.size)

x = $c['atoi', 'IS']['100']
test("atoi('100')", x[0], 100)

x = $c['atoi', 'IS']['-100']
test("atoi('-100')", x[0], -100)

x = $c['atol', 'LS']['100']
test("atol('100')", x[0], 100)

x = $c['atol', 'LS']['-100']
test("atol('-100')", x[0], -100)

x = $c['atof', 'DS']['0.1']
test("atof('0.1')", x[0], 0.09..0.11)

x = $c['atof', 'DS']['-0.1']
test("atof('-0.1')", x[0], -0.11..-0.09)

buff = "xxxx"
str  = "abc"
x = $c['strcpy', 'PsS'][buff,str]
test("strcpy('xxxx','abc')", buff, "abc\0")
test("strcpy('xxxx','abc')", x[0].to_s(4), "abc\0")

buff = "xxxx"
str  = "abc"
x = $c['strncpy', 'SsSI'][buff,str,3]
test("strncpy('xxxx','abc',3)", buff, "abcx")
test("strncpy('xxxx','abc',3)", x[0], "abcx")

buff = "foobarbaz"
cb = DL.set_callback(DL::TYPE_INT){|x,y,*rest| DL::dlwrap(x)[0] <=> DL::dlwrap(y)[0]}
x = $c['qsort', '0PIIP'][buff,buff.size,1,PtrData.new(cb)]
test("qsort('foobarbaz',...)", buff, 'aabbfoorz')

print("#{$num} test(s), #{$err} fail(s)\n")
