require 'dl'
require 'dl/callback'

LIBC_SO = ARGV[0] || "/lib/libc.so.6" # cygwin1.dll
LIBM_SO = ARGV[1] || "/lib/libm.so.6" # cygwin1.dll


$c = DL.dlopen(LIBC_SO)
$m = DL.dlopen(LIBM_SO)
$num = 0
$err = 0

def test(phase, x1, x2)
  $num += 1
  print(phase, ": ")
  if( x2 === x1 )
    print("ok\n")
  else
    print("fail\n")
    $err += 1
  end
end

include DL
include Math

x = CFunc.new($m['sin'], DL::TYPE_DOUBLE, 'sin').call([3.14/2].pack("d").unpack("l!*"))
test("sin(3.14/2)", x, Math.sin(3.14/2))

x = CFunc.new($m['sin'], DL::TYPE_DOUBLE, 'sin').call([-3.14/2].pack("d").unpack("l!*"))
test("sin(-3.14/2)", x, Math.sin(-3.14/2))

x = CFunc.new($c['strlen'], DL::TYPE_INT, 'strlen').call(["abc"].pack("p").unpack("l!*"))
test("strlen('abc')", x, "abc".size)

x = CFunc.new($c['atoi'], DL::TYPE_INT, 'atoi').call(["100"].pack("p").unpack("l!*"))
test("atoi('100')", x, 100)

x = CFunc.new($c['atoi'], DL::TYPE_INT, 'atoi').call(["-100"].pack("p").unpack("l!*"))
test("atoi('-100')", x, -100)

x = CFunc.new($c['atol'], DL::TYPE_LONG, 'atol').call(["100"].pack("p").unpack("l!*"))
test("atol('100')", x, 100)

x = CFunc.new($c['atol'], DL::TYPE_LONG, 'atol').call(["-100"].pack("p").unpack("l!*"))
test("atol('-100')", x, -100)

x = CFunc.new($c['atof'], DL::TYPE_DOUBLE, 'atof').call(["0.1"].pack("p").unpack("l!*"))
test("atof('0.1')", x, 0.09..0.11)

x = CFunc.new($c['atof'], DL::TYPE_DOUBLE, 'atof').call(["-0.1"].pack("p").unpack("l!*"))
test("atof('-0.1')", x, -0.11 .. -0.09)

buff = "xxxx"
str  = "abc"
x = CFunc.new($c['strcpy'], DL::TYPE_VOIDP, 'strcpy').call([buff,str].pack("pp").unpack("l!*"))
test("strcpy('xxxx','abc')", buff, "abc\0")
test("strcpy('xxxx','abc')", DL.dlwrap(x,4), "abc\0")

buff = "xxxx"
str  = "abc"
x = CFunc.new($c['strncpy'], DL::TYPE_VOIDP, 'strncpy').call([buff,str,3].pack("ppi").unpack("l!*"))
test("strncpy('xxxx','abc',3)", buff, "abcx")
test("strncpy('xxxx','abc',3)", DL.dlwrap(x,4), "abcx")

buff = "foobarbaz"
cb = set_callback(DL::TYPE_INT){|x,y,*rest| DL::dlwrap(x)[0] <=> DL::dlwrap(y)[0]}
x = CFunc.new($c['qsort'], DL::TYPE_VOID, 'qsort').call([buff, buff.size, 1, cb].pack("pI!I!L!").unpack("l!*"))
test("qsort('foobarbaz',...)", buff, 'aabbfoorz')

print("#{$num} test(s), #{$err} fail(s)\n")
